package com.tzld.piaoquan.ad.engine.service.score.strategy;

import com.alibaba.fastjson.JSONObject;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.ad.engine.commons.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.commons.util.DateUtils;
import com.tzld.piaoquan.ad.engine.commons.util.ExtractorUtils;
import com.tzld.piaoquan.ad.engine.commons.util.NumUtil;
import com.tzld.piaoquan.ad.engine.commons.util.SimilarityUtils;
import com.tzld.piaoquan.ad.engine.service.entity.FeatureContainer;
import com.tzld.piaoquan.ad.engine.service.entity.GuaranteeView;
import com.tzld.piaoquan.ad.engine.commons.enums.RedisPrefixEnum;
import com.tzld.piaoquan.ad.engine.commons.param.RankRecommendRequestParam;
import com.tzld.piaoquan.ad.engine.commons.redis.AdRedisHelper;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.service.feature.Feature;
import com.tzld.piaoquan.ad.engine.service.feature.FeatureService;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.xm.Similarity;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import static com.tzld.piaoquan.ad.engine.commons.math.Const.*;

@Slf4j
public abstract class RankStrategyBasic implements RankStrategy {

    @ApolloJsonValue("${no.postback.conversion.adverids:[]}")
    private Set<String> noPostbackConversionAdVerIds;

    @ApolloJsonValue("${creative.model.score.coefficient:{}}")
    private Map<Long, Double> creativeScoreCoefficient;

    @Value("${guarantee.exp:742}")
    protected String guaranteeExp;
    @ApolloJsonValue("${alpha:1.0}")
    protected Double alpha;

    @Value("${word2vec.exp:694}")
    protected String word2vecExp;

    @Autowired
    private FeatureService featureService;
    @Autowired
    protected AdRedisHelper adRedisHelper;

    String adPlatformGuaranteeKey = "ad:platform:guarantee:data:{date}:{adverId}";

    protected static final List<String> hasChannelScenes = new ArrayList<String>() {{
        add("DaiTou");
        add("GzhTouLiu");
        add("daitou");
        add("dyyjs");
        add("dyyqw");
        add("fwhdyy");
        add("gzhhz");
        add("gzhhzzx");
        add("shequn");
        add("touliu");
        add("xcxdt");
    }};


    protected Feature getFeature(ScoreParam param, RankRecommendRequestParam request) {
        List<AdPlatformCreativeDTO> adIdList = request.getAdIdList();
        List<String> cidList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getCreativeId)
                .map(Object::toString)
                .collect(Collectors.toList());

        List<String> adVerIdList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getAdVerId)
                .filter(StringUtils::isNotBlank)
                .distinct()
                .collect(Collectors.toList());

        List<Long> skuIdList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getSkuId)
                .filter(Objects::nonNull)
                .distinct()
                .collect(Collectors.toList());
        return featureService.getFeature(cidList, adVerIdList, skuIdList, param);
    }

    protected Set<String> getNoApiAdVerIds() {
        Set<String> noApiAdVerIds = adRedisHelper.sMembers(RedisPrefixEnum.ADVER_IS_API_EQ_0_IDS.getPrefix());
        log.info("Redis中缓存的没有回传广告主ID集合: {}", noApiAdVerIds);
        if (CollectionUtils.isEmpty(noApiAdVerIds)) {
            noApiAdVerIds = noPostbackConversionAdVerIds;
        }
        return noApiAdVerIds;
    }

    protected Map<Long, Double> getCreativeScoreCoefficient() {
        if (Objects.isNull(creativeScoreCoefficient)) {
            return new HashMap<>();
        }
        return creativeScoreCoefficient;
    }

    protected Map<String, String> getReqFeature(ScoreParam scoreParam, RankRecommendRequestParam request) {
        Map<String, String> map = new HashMap<>();
        map.put("brand", scoreParam.getRequestContext().getMachineinfoBrand().equalsIgnoreCase("-1") ?
                "" : scoreParam.getRequestContext().getMachineinfoBrand().toUpperCase());
        map.put("region", Objects.equals(scoreParam.getRequestContext().getRegion(), "-1") ?
                "" : scoreParam.getRequestContext().getRegion());
        map.put("city", Objects.equals(scoreParam.getRequestContext().getCity(), "-1") ?
                "" : scoreParam.getRequestContext().getCity());
        map.put("vid", String.valueOf(request.getVideoId()));
        map.put("apptype", String.valueOf(request.getAppType()));
        String rootSessionId = request.getRootSessionId();
        String sessionId = request.getSessionId();
        String subSessionId = request.getSubSessionId();
        String rootSourceId = request.getRootSourceId();
        int isFirstLayer;
        if (Objects.equals(rootSessionId, subSessionId) || Objects.equals(rootSessionId, sessionId)) {
            isFirstLayer = 1;
        } else {
            isFirstLayer = 2;
        }
        map.put("is_first_layer", String.valueOf(isFirstLayer));
        if (StringUtils.isNotEmpty(rootSourceId)) {
            String rootSourceScene = rootSourceId.split("_")[0];
            map.put("root_source_scene", rootSourceScene);
            if (hasChannelScenes.contains(rootSourceScene) && rootSourceId.split("_").length > 1) {
                String rootSourceChannel = rootSourceId.split("_")[1];
                map.put("root_source_channel", rootSourceChannel);
            }
        }
        return map;
    }

    protected Map<String, String> handleSceneFeature(long ts) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        sceneFeatureMap.put("hour_" + DateUtils.getHourByTimestamp(ts), "0.1");
        sceneFeatureMap.put("dayofweek_" + DateUtils.getDayOrWeekByTimestamp(ts), "0.1");
        sceneFeatureMap.put("hour", String.valueOf(DateUtils.getHourByTimestamp(ts)));
        sceneFeatureMap.put("hour_quarter", String.valueOf(DateUtils.getHourQuarter(ts)));
        sceneFeatureMap.put("ts", String.valueOf(ts));
        return sceneFeatureMap;
    }

    protected FeatureContainer extractAndProcessFeatures(
            ScoreParam scoreParam,
            RankRecommendRequestParam request,
            long ts) {
        FeatureContainer container = new FeatureContainer();
        Feature feature = getFeature(scoreParam, request);
        // 基础特征
        container.setUserFeature(feature.getUserFeature());
        container.setVideoFeature(feature.getVideoFeature());
        container.setAdVerFeature(feature.getAdVerFeature());
        container.setCidFeature(feature.getCidFeature());
        container.setSkuFeature(feature.getSkuFeature());

        // 处理后的特征
        container.setReqFeature(getReqFeature(scoreParam, request));
        container.setSceneFeatureMap(handleSceneFeature(ts));

        // 子特征
        container.setC1Feature(container.getUserFeature()
                .getOrDefault("alg_mid_feature_ad_action", new HashMap<>()));
        container.setD2Feature(container.getVideoFeature()
                .getOrDefault("alg_cid_feature_vid_cf_rank", new HashMap<>()));
        container.setD3Feature(container.getVideoFeature()
                .getOrDefault("alg_vid_feature_basic_info", new HashMap<>()));
        container.setE1Feature(container.getUserFeature()
                .getOrDefault("alg_mid_feature_return_tags", new HashMap<>()));
        container.setE2Feature(container.getUserFeature()
                .getOrDefault("alg_mid_feature_share_tags", new HashMap<>()));
        container.setG1Feature(container.getUserFeature()
                .getOrDefault("mid_return_video_cate", new HashMap<>()));
        container.setG2Feature(container.getUserFeature()
                .getOrDefault("mid_share_video_cate", new HashMap<>()));

        return container;
    }

    protected static class Tuple5 {
        public String f1;
        public String f2;
        public String f3;
        public String f4;
        public String f5;

        public Tuple5(String f1, String f2, String f3, String f4, String f5) {
            this.f1 = f1;
            this.f2 = f2;
            this.f3 = f3;
            this.f4 = f4;
            this.f5 = f5;
        }
    }

    protected static class TupleMapEntry<T> {
        public String key;
        public T value;

        public TupleMapEntry(String key, T value) {
            this.key = key;
            this.value = value;
        }
    }

    protected static class Tuple2<F1, F2> {
        public F1 f1;

        public F2 f2;

        public Tuple2(F1 first, F2 name) {
            this.f1 = first;
            this.f2 = name;
        }

    }

    protected Map<String, GuaranteeView> getGuaranteeViewMap(RankRecommendRequestParam request, ScoreParam scoreParam) {
        Map<String, GuaranteeView> map = new HashMap<>();
        try {
            String thatDayDateString = DateUtils.getThatDayDateString();
            String redisKey = adPlatformGuaranteeKey.replace("{date}", thatDayDateString);
            List<String> adVerIds = request.getAdIdList().stream().map(AdPlatformCreativeDTO::getAdVerId).distinct()
                    .filter(Objects::nonNull).collect(Collectors.toList());
            List<String> redisKeys = adVerIds.stream().map(e -> redisKey.replace("{adverId}", e)).collect(Collectors.toList());
            List<String> values = adRedisHelper.mget(redisKeys);
            if (CollectionUtils.isNotEmpty(values)) {
                for (int i = 0; i < redisKeys.size(); i++) {
                    String value = values.get(i);
                    if (value != null) {
                        GuaranteeView guaranteeView = JSONObject.parseObject(value, GuaranteeView.class);
                        map.put(adVerIds.get(i), guaranteeView);
                    }
                }
            }
        } catch (Exception e) {
            log.error("getGuaranteeViewMap error", e);
        }

        return map;
    }

    protected void setGuaranteeWeight(Map<String, GuaranteeView> map, String adVerId, Map<String, Object> ext) {
        if (MapUtils.isNotEmpty(map)) {
            GuaranteeView guaranteeView = map.get(adVerId);
            if (guaranteeView != null) {
                double guaranteeWeight = calculateGuaranteedWeight(guaranteeView);
                boolean isGuaranteed = isGuaranteed(guaranteeView);
                ext.put("guaranteeView", guaranteeView.toString());
                ext.put("guaranteeWeight", guaranteeWeight);
                ext.put("isGuaranteed", isGuaranteed);
            }
        }
    }

    protected double calculateGuaranteedWeight(GuaranteeView guaranteeView) {
        if (guaranteeView == null) {
            return 1.0;
        }
        double guaranteeWeight;
        if (guaranteeView.getGuaranteeNum() == null || guaranteeView.getGuaranteeNum() == 0
                || guaranteeView.getGuaranteeRate() == null || guaranteeView.getGuaranteeRate() == 0.0) {
            guaranteeWeight = 1.0;
        } else {
            if (guaranteeView.getAdrAlgoViewNum() != null &&
                    guaranteeView.getGuaranteeNum() <= guaranteeView.getAdrAlgoViewNum()) {
                guaranteeWeight = 0.0;
            } else {
                int allViewNum = guaranteeView.getAllAlgoViewNum() == null || guaranteeView.getAllAlgoViewNum() == 0 ?
                        1 : guaranteeView.getAllAlgoViewNum();
                int adrAlogViewNum = guaranteeView.getAdrAlgoViewNum() == null || guaranteeView.getAdrAlgoViewNum() == 0 ?
                        1 : guaranteeView.getAdrAlgoViewNum();
                //guaranteeView.getGuaranteeRate() 是百分之几  要乘0.01
                guaranteeWeight = guaranteeView.getGuaranteeRate() * 0.01 * allViewNum / adrAlogViewNum;
                if (guaranteeWeight < 0.5) {
                    guaranteeWeight = 0.5;
                }
                if (guaranteeWeight > 2) {
                    guaranteeWeight = 2;
                }
            }
        }
        return guaranteeWeight;
    }

    protected boolean isGuaranteed(GuaranteeView guaranteeView) {
        if (guaranteeView == null) {
            return false;
        }
        return guaranteeView.getGuaranteeNum() != null && guaranteeView.getGuaranteeNum() != 0
                && guaranteeView.getGuaranteeRate() != null && guaranteeView.getGuaranteeRate() != 0.0;
    }

    protected double getGuaranteeScoreCoefficient(ScoreParam scoreParam, Map<String, Object> ext) {
        if (ext.get("guaranteeWeight") != null && (double) ext.get("guaranteeWeight") == 0) {
            return 0.0;
        }
        if (scoreParam.getExpCodeSet().contains(guaranteeExp)) {
            if (ext.get("guaranteeWeight") == null) {
                return 1.0;
            } else {
                return Math.pow((double) ext.get("guaranteeWeight"), alpha);
            }
        } else {
            return 1.0;
        }
    }

    protected void handleB1Feature(Map<String, String> b1Feature, Map<String, String> cidFeatureMap, String cid) {
        cidFeatureMap.put("cid_" + cid, "0.1");
        if (StringUtils.isNotBlank(b1Feature.get("adverid"))) {
            String adVerId = b1Feature.get("adverid");
            cidFeatureMap.put("adverid_" + adVerId, "0.1");
        }
        if (StringUtils.isNotBlank(b1Feature.get("cpa"))) {
            String cpa = b1Feature.get("cpa");
            cidFeatureMap.put("cpa", cpa);
        }
    }

    protected void handleB2ToB5AndB8ToB9Feature(Map<String, Map<String, String>> c1Feature, Map<String, Map<String, String>> adVerFeature, Map<String, String> cidFeatureMap) {
        Map<String, String> b2Feature = adVerFeature.getOrDefault("alg_cid_feature_adver_action", new HashMap<>());
        Map<String, String> b3Feature = c1Feature.getOrDefault("alg_cid_feature_cid_action", new HashMap<>());
        Map<String, String> b4Feature = c1Feature.getOrDefault("alg_cid_feature_region_action", new HashMap<>());
        Map<String, String> b5Feature = c1Feature.getOrDefault("alg_cid_feature_app_action", new HashMap<>());
        Map<String, String> b8Feature = c1Feature.getOrDefault("alg_cid_feature_brand_action", new HashMap<>());
        Map<String, String> b9Feature = c1Feature.getOrDefault("alg_cid_feature_weChatVersion_action", new HashMap<>());

        List<String> timeList = Arrays.asList("1h", "2h", "3h", "6h", "12h", "1d", "3d", "7d", "yesterday", "today");
        List<Tuple2<Map<String, String>, String>> featureList = Arrays.asList(
                new Tuple2<>(b2Feature, "b2"),
                new Tuple2<>(b3Feature, "b3"),
                new Tuple2<>(b4Feature, "b4"),
                new Tuple2<>(b5Feature, "b5"),
                new Tuple2<>(b8Feature, "b8"),
                new Tuple2<>(b9Feature, "b9")
        );
        for (Tuple2<Map<String, String>, String> tuple2 : featureList) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            for (String time : timeList) {
                double view = Double.parseDouble(feature.getOrDefault("ad_view_" + time, "0"));
                double click = Double.parseDouble(feature.getOrDefault("ad_click_" + time, "0"));
                double conver = Double.parseDouble(feature.getOrDefault("ad_conversion_" + time, "0"));
                double income = Double.parseDouble(feature.getOrDefault("ad_income_" + time, "0"));
                double cpc = NumUtil.div(income, click);
                double ctr = NumUtil.divSmoothV2(click, view, CTR_SMOOTH_BETA_FACTOR);
                double ctcvr = NumUtil.divSmoothV2(conver, view, CTCVR_SMOOTH_BETA_FACTOR);
                double ecpm = ctr * cpc * 1000;
                cidFeatureMap.put(prefix + "_" + time + "_ctr", String.valueOf(ctr));
                cidFeatureMap.put(prefix + "_" + time + "_ctcvr", String.valueOf(ctcvr));
                cidFeatureMap.put(prefix + "_" + time + "_cvr", String.valueOf(NumUtil.divSmoothV2(conver, click, CVR_SMOOTH_BETA_FACTOR)));
                cidFeatureMap.put(prefix + "_" + time + "_conver", String.valueOf(conver));
                cidFeatureMap.put(prefix + "_" + time + "_ecpm", String.valueOf(ecpm));

                cidFeatureMap.put(prefix + "_" + time + "_click", String.valueOf(click));
                cidFeatureMap.put(prefix + "_" + time + "_conver*log(view)", String.valueOf(conver * NumUtil.log(view)));
                cidFeatureMap.put(prefix + "_" + time + "_conver*ctcvr", String.valueOf(conver * ctcvr));
            }
        }

    }

    protected void handleB6ToB7Feature(Map<String, Map<String, String>> c1Feature, Map<String, String> cidFeatureMap) {
        Map<String, String> b6Feature = c1Feature.getOrDefault("alg_cid_feature_week_action", new HashMap<>());
        Map<String, String> b7Feature = c1Feature.getOrDefault("alg_cid_feature_hour_action", new HashMap<>());

        List<String> timeList = Arrays.asList("7d", "14d");
        List<Tuple2<Map<String, String>, String>> featureList = Arrays.asList(
                new Tuple2<>(b6Feature, "b6"),
                new Tuple2<>(b7Feature, "b7")
        );
        for (Tuple2<Map<String, String>, String> tuple2 : featureList) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            for (String time : timeList) {
                double view = Double.parseDouble(feature.getOrDefault("ad_view_" + time, "0"));
                double click = Double.parseDouble(feature.getOrDefault("ad_click_" + time, "0"));
                double conver = Double.parseDouble(feature.getOrDefault("ad_conversion_" + time, "0"));
                double income = Double.parseDouble(feature.getOrDefault("ad_income_" + time, "0"));
                double cpc = NumUtil.div(income, click);
                double ctr = NumUtil.divSmoothV2(click, view, CTR_SMOOTH_BETA_FACTOR);
                double ctcvr = NumUtil.divSmoothV2(conver, view, CTCVR_SMOOTH_BETA_FACTOR);
                double ecpm = ctr * cpc * 1000;
                cidFeatureMap.put(prefix + "_" + time + "_ctr", String.valueOf(ctr));
                cidFeatureMap.put(prefix + "_" + time + "_ctcvr", String.valueOf(ctcvr));
                cidFeatureMap.put(prefix + "_" + time + "_cvr", String.valueOf(NumUtil.divSmoothV2(conver, click, CVR_SMOOTH_BETA_FACTOR)));
                cidFeatureMap.put(prefix + "_" + time + "_conver", String.valueOf(conver));
                cidFeatureMap.put(prefix + "_" + time + "_ecpm", String.valueOf(ecpm));

                cidFeatureMap.put(prefix + "_" + time + "_click", String.valueOf(click));
                cidFeatureMap.put(prefix + "_" + time + "_conver*log(view)", String.valueOf(conver * NumUtil.log(view)));
                cidFeatureMap.put(prefix + "_" + time + "_conver*ctcvr", String.valueOf(conver * ctcvr));
            }
        }

    }

    protected List<TupleMapEntry<Tuple5>> handleC1Feature(Map<String, String> c1Feature, Map<String, String> featureMap) {

        // 用户特征
        List<TupleMapEntry<Tuple5>> midActionList = new ArrayList<>();
        if (c1Feature.containsKey("action")) {
            String action = c1Feature.get("action");
            midActionList = Arrays.stream(action.split(","))
                    .map(r -> {
                        String[] rList = r.split(":");
                        Tuple5 tuple5 = new Tuple5(rList[1], rList[2], rList[3], rList[4], rList[5]);
                        return new TupleMapEntry<>(rList[0], tuple5);
                    })
                    // TODO 倒排
                    .sorted((a, b) -> Integer.compare(Integer.parseInt(b.value.f1), Integer.parseInt(a.value.f1)))
                    .collect(Collectors.toList());
        }

        double viewAll = midActionList.size();
        double clickAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f2)).sum();
        double converAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f3)).sum();
        double incomeAll = midActionList.stream().mapToInt(e -> Integer.parseInt(e.value.f4)).sum();
        featureMap.put("viewAll", String.valueOf(viewAll));
        featureMap.put("clickAll", String.valueOf(clickAll));
        featureMap.put("converAll", String.valueOf(converAll));
        featureMap.put("incomeAll", String.valueOf(incomeAll));
        featureMap.put("ctr_all", String.valueOf(NumUtil.div(clickAll, viewAll)));
        featureMap.put("ctcvr_all", String.valueOf(NumUtil.div(converAll, viewAll)));
        featureMap.put("cvr_all", String.valueOf(NumUtil.div(clickAll, converAll)));
        featureMap.put("ecpm_all", String.valueOf(NumUtil.div(incomeAll * 1000, viewAll)));

        return midActionList;
    }

    protected void handleC1UIFeature(Map<String, Double> midTimeDiffMap, Map<String, Double> midActionStatic, Map<String, String> featureMap, String cid) {
        if (midTimeDiffMap.containsKey("timediff_view_" + cid)) {
            featureMap.put("timediff_view", String.valueOf(midTimeDiffMap.getOrDefault("timediff_view_" + cid, 0.0)));
        }
        if (midTimeDiffMap.containsKey("timediff_click_" + cid)) {
            featureMap.put("timediff_click", String.valueOf(midTimeDiffMap.getOrDefault("timediff_click_" + cid, 0.0)));
        }
        if (midTimeDiffMap.containsKey("timediff_conver_" + cid)) {
            featureMap.put("timediff_conver", String.valueOf(midTimeDiffMap.getOrDefault("timediff_conver_" + cid, 0.0)));
        }
        if (midActionStatic.containsKey("actionstatic_view_" + cid)) {
            featureMap.put("actionstatic_view", String.valueOf(midActionStatic.getOrDefault("actionstatic_view_" + cid, 0.0)));
        }
        if (midActionStatic.containsKey("actionstatic_click_" + cid)) {
            featureMap.put("actionstatic_click", String.valueOf(midActionStatic.getOrDefault("actionstatic_click_" + cid, 0.0)));
        }
        if (midActionStatic.containsKey("actionstatic_conver_" + cid)) {
            featureMap.put("actionstatic_conver", String.valueOf(midActionStatic.getOrDefault("actionstatic_conver_" + cid, 0.0)));
        }
        if (midActionStatic.containsKey("actionstatic_income_" + cid)) {
            featureMap.put("actionstatic_income", String.valueOf(midActionStatic.getOrDefault("actionstatic_income_" + cid, 0.0)));
        }
        if (midActionStatic.containsKey("actionstatic_view_" + cid) && midActionStatic.containsKey("actionstatic_click_" + cid)) {
            double ctr = NumUtil.div(
                    midActionStatic.getOrDefault("actionstatic_click_" + cid, 0.0),
                    midActionStatic.getOrDefault("actionstatic_view_" + cid, 0.0)
            );
            featureMap.put("actionstatic_ctr", String.valueOf(ctr));
        }
        if (midActionStatic.containsKey("actionstatic_view_" + cid) && midActionStatic.containsKey("actionstatic_conver_" + cid)) {
            double ctcvr = NumUtil.div(midActionStatic.getOrDefault("actionstatic_conver_" + cid, 0.0), midActionStatic.getOrDefault("actionstatic_view_" + cid, 0.0));
            featureMap.put("actionstatic_ctcvr", String.valueOf(ctcvr));
        }
        if (midActionStatic.containsKey("actionstatic_conver_" + cid) && midActionStatic.containsKey("actionstatic_click_" + cid)) {
            double cvr = NumUtil.div(midActionStatic.getOrDefault("actionstatic_conver_" + cid, 0.0), midActionStatic.getOrDefault("actionstatic_click_" + cid, 0.0));
            featureMap.put("actionstatic_cvr", String.valueOf(cvr));
        }
    }

    protected void handleD1Feature(Map<String, String> d1Feature, Map<String, String> featureMap) {
        for (String prefix : Arrays.asList("3h", "6h", "12h", "1d", "3d", "7d")) {
            double view = Double.parseDouble(d1Feature.getOrDefault("ad_view_" + prefix, "0"));
            double click = Double.parseDouble(d1Feature.getOrDefault("ad_click_" + prefix, "0"));
            double conver = Double.parseDouble(d1Feature.getOrDefault("ad_conversion_" + prefix, "0"));
            double income = Double.parseDouble(d1Feature.getOrDefault("ad_income_" + prefix, "0"));
            double cpc = NumUtil.div(income, click);
            double ctr = NumUtil.divSmoothV2(click, view, CTR_SMOOTH_BETA_FACTOR);
            featureMap.put("d1_feature_" + prefix + "_ctr", String.valueOf(ctr));
            featureMap.put("d1_feature_" + prefix + "_ctcvr", String.valueOf(NumUtil.divSmoothV2(conver, view, CTCVR_SMOOTH_BETA_FACTOR)));
            featureMap.put("d1_feature_" + prefix + "_cvr", String.valueOf(NumUtil.divSmoothV2(conver, click, CVR_SMOOTH_BETA_FACTOR)));
            featureMap.put("d1_feature_" + prefix + "_conver", String.valueOf(conver));
            featureMap.put("d1_feature_" + prefix + "_ecpm", String.valueOf(ctr * cpc * 1000));
        }
    }

    protected void handleD2Feature(Map<String, Map<String, Double>> vidRankMaps, Map<String, String> featureMap, String cid) {
        if (MapUtils.isEmpty(vidRankMaps)) {
            return;
        }

        List<String> prefixes1 = Arrays.asList("ctr", "ctcvr", "ecpm");
        // List<String> prefixes1 = Arrays.asList("ctr", "ctcvr");
        List<String> prefixes2 = Arrays.asList("1d", "3d", "7d", "14d");

        for (String prefix1 : prefixes1) {
            for (String prefix2 : prefixes2) {
                String combinedKey = prefix1 + "_" + prefix2;
                if (vidRankMaps.containsKey(combinedKey)) {
                    Double rank = vidRankMaps.get(combinedKey).getOrDefault(cid, 0.0);
                    if (rank >= 1.0) {
                        featureMap.put("vid_rank_" + combinedKey, String.valueOf(NumUtil.div(1, rank)));
                    }
                }
            }
        }
    }

    protected void handleD3AndB1Feature(Map<String, String> d3Feature, String cTitle, Map<String, String> featureMap,
                                        ScoreParam scoreParam) {
        if (MapUtils.isEmpty(d3Feature) || !d3Feature.containsKey("title") || StringUtils.isEmpty(cTitle)) {
            return;
        }
        String vTitle = d3Feature.get("title");
        double score;
        if (scoreParam.getExpCodeSet().contains(word2vecExp)) {
            score = SimilarityUtils.word2VecSimilarity(cTitle, vTitle);
        } else {
            score = Similarity.conceptSimilarity(cTitle, vTitle);
        }
        featureMap.put("ctitle_vtitle_similarity", String.valueOf(score));
    }

    protected void handleE1AndE2Feature(Map<String, String> e1Feature, Map<String, String> e2Feature, String title,
                                        Map<String, String> featureMap, ScoreParam scoreParam) {
        if (StringUtils.isEmpty(title)) {
            return;
        }

        List<Tuple2<Map<String, String>, String>> tuple2List = Arrays.asList(new Tuple2<>(e1Feature, "e1"), new Tuple2<>(e2Feature, "e2"));

        List<String> tagsFieldList = Arrays.asList("tags_3d", "tags_7d", "tags_14d");
        for (Tuple2<Map<String, String>, String> tuple2 : tuple2List) {
            Map<String, String> feature = tuple2.f1;
            String prefix = tuple2.f2;
            if (MapUtils.isEmpty(feature)) {
                continue;
            }

            for (String tagsField : tagsFieldList) {
                if (StringUtils.isNotEmpty(feature.get(tagsField))) {
                    String tags = feature.get(tagsField);
                    // Double[] doubles = ExtractorUtils.funcC34567ForTags(tags, title);
                    Double[] doubles;
                    if (scoreParam.getExpCodeSet().contains(word2vecExp)) {
                        doubles = ExtractorUtils.funcC34567ForTagsNew(tags, title);
                    } else {
                        doubles = ExtractorUtils.funcC34567ForTags(tags, title);
                    }
                    featureMap.put(prefix + "_" + tagsField + "_matchnum", String.valueOf(doubles[0]));
                    featureMap.put(prefix + "_" + tagsField + "_maxscore", String.valueOf(doubles[1]));
                    featureMap.put(prefix + "_" + tagsField + "_avgscore", String.valueOf(doubles[2]));
                }
            }
        }
    }

    protected Map<String, Double> parseC1FeatureListToTimeDiffMap(List<TupleMapEntry<Tuple5>> midActionList, long ts) {
        Map<String, Double> midTimeDiffMap = new HashMap<>();
        for (TupleMapEntry<Tuple5> entry : midActionList) {
            String cid = entry.key;
            double tsHistory = Double.parseDouble(entry.value.f1);
            double click = Double.parseDouble(entry.value.f2);
            double conver = Double.parseDouble(entry.value.f3);
            double d = (ts - tsHistory) / 3600 / 24;
            if (!midTimeDiffMap.containsKey("timediff_view_" + cid)) {
                midTimeDiffMap.put("timediff_view_" + cid, NumUtil.div(1, d));
            }
            if (!midTimeDiffMap.containsKey("timediff_click_" + cid) && click > 0) {
                midTimeDiffMap.put("timediff_click_" + cid, NumUtil.div(1, d));
            }
            if (!midTimeDiffMap.containsKey("timediff_conver_" + cid) && conver > 0) {
                midTimeDiffMap.put("timediff_conver_" + cid, NumUtil.div(1, d));
            }
        }
        return midTimeDiffMap;
    }

    protected Map<String, Double> parseC1FeatureListToActionStaticMap(List<TupleMapEntry<Tuple5>> midActionList) {
        Map<String, Double> midActionStaticsMap = new HashMap<>();
        for (TupleMapEntry<Tuple5> entry : midActionList) {
            String cid = entry.key;
            double click = Double.parseDouble(entry.value.f2);
            double conver = Double.parseDouble(entry.value.f3);
            double income = Double.parseDouble(entry.value.f4);

            Double viewSum = midActionStaticsMap.getOrDefault("actionstatic_view_" + cid, 0.0);
            midActionStaticsMap.put("actionstatic_view_" + cid, 1 + viewSum);

            Double clickSum = midActionStaticsMap.getOrDefault("actionstatic_click_" + cid, 0.0);
            midActionStaticsMap.put("actionstatic_click_" + cid, clickSum + click);

            Double converSum = midActionStaticsMap.getOrDefault("actionstatic_conver_" + cid, 0.0);
            midActionStaticsMap.put("actionstatic_conver_" + cid, converSum + conver);

            Double incomSum = midActionStaticsMap.getOrDefault("actionstatic_income_" + cid, 0.0);
            midActionStaticsMap.put("actionstatic_income_" + cid, incomSum + income);
        }

        return midActionStaticsMap;
    }

    protected Map<String, Map<String, Double>> parseD2FeatureMap(Map<String, String> d2Feature) {
        Map<String, Map<String, Double>> vidRankMaps = new HashMap<>();
        for (Map.Entry<String, String> entry : d2Feature.entrySet()) {
            String key = entry.getKey();
            String value = entry.getValue();
            Map<String, Double> valueMap = Arrays.stream(value.split(",")).map(r -> r.split(":")).collect(Collectors.toMap(rList -> rList[0], rList -> Double.parseDouble(rList[2])));
            vidRankMaps.put(key, valueMap);
        }
        return vidRankMaps;
    }

    protected void setMetaFeatureMap(FeatureContainer featureContainer, AdRankItem adRankItem) {
        Map<String, Map<String, String>> userFeature = featureContainer.getUserFeature();
        Map<String, Map<String, String>> videoFeature = featureContainer.getVideoFeature();
        Map<String, Map<String, Map<String, String>>> allAdVerFeature = featureContainer.getAdVerFeature();
        Map<String, Map<String, Map<String, String>>> allCidFeature = featureContainer.getCidFeature();
        Map<String, Map<String, Map<String, String>>> allSkuFeature = featureContainer.getSkuFeature();
        for (Map.Entry<String, Map<String, String>> entry : videoFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        for (Map.Entry<String, Map<String, String>> entry : userFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> adVerFeature = allAdVerFeature.getOrDefault(adRankItem.getAdVerId(), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : adVerFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> cidFeature = allCidFeature.getOrDefault(String.valueOf(adRankItem.getAdId()), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : cidFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> skuFeature = allSkuFeature.getOrDefault(String.valueOf(adRankItem.getSkuId()), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : skuFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }
        adRankItem.getMetaFeatureMap().put("reqFeature", featureContainer.getReqFeature());
        adRankItem.getMetaFeatureMap().put("sceneFeature", featureContainer.getSceneFeatureMap());
    }


}
